﻿//----------------------------------------------------------------- 2013-11-08 ------------
// 電力計 計測器版　液晶／無線通信共通ファームウェア
//
//  ハードウェアバージョンの違いによる相違箇所は主要な物だけ #ifdef で分岐させている
//　そのため不要な物も副作用の無い範囲で定義は残っている
//
// 要変更箇所：
//  ・HW_VERSION1/2/3
//  ・FREQ 50/60
//-------------------------------------------------------- written by iizuka --------------
#include <SPI.h>

// #define DEBUG

// #define HW_VERSION1     // ハードウェアのバージョンによりここを切り替える
#define HW_VERSION2     // 
// #define HW_VERSION3     // 

#ifdef HW_VERSION3
  #include <Mirf.h>                      // Version3 は無線モジュール
  #include <MirfHardwareSpiDriver.h>
#else
  #include <LiquidCrystal.h>             // Version1,2 は液晶
  LiquidCrystal lcd(5,6,7,8,9,10);
#endif

#ifdef HW_VERSION1
  #define VERSION_NUMBER  '1'      // HardWare Version
#endif
#ifdef HW_VERSION2
  #define VERSION_NUMBER  '2'      // HardWare Version
#endif
#ifdef HW_VERSION3
  #define VERSION_NUMBER  '3'      // HardWare Version
#endif

#define FW_VERSION  "1.0.01"       // FirmWare Version （必ず6文字にすること）

//---------------- 環境により変更する項目 -------------------------------------------
#define FREQ    50   // 50Hz
//  #define FREQ 60   // 60Hz
#define CT_TURN                   3000     // ＣＴセンサーのコイル巻き数 CTL-10-CLS は 3000巻き
#define CH1_R_LOW_RESOLUTION       200     // 回路図のR1(V3:R8 )=200Ω  抵抗は出力が振り切れない範囲で大きい方が計測精度が良い
#define CH1_R_HIGH_RESOLUTION     3000     // 回路図のR3(V3:R10)=3kΩ   実際の抵抗は誤差があるのでテスターで実測して
#define CH2_R_LOW_RESOLUTION       200     // 回路図のR2(V3:R9 )=200Ω  この値を変更すると精度が向上する
#define CH2_R_HIGH_RESOLUTION     3000     // 回路図のR4(V3:R11)=3kΩ
#define SAMPLES 120    // 1サイクル(1/50秒)に何回サンプリングするか
                       // 大きいほど精度が高く計測できるが大きすぎるとＡ／Ｄ変換の速度が間に合わない 150あたりが限界
#define sampling_interval 1000             // サンプリング間隔[msec] 最大65秒まで可
                                           // 但し１分以上にするとグラフ表示プログラムのデータが歯抜けになるため３０秒以下推奨
#define EEP_MAX_ADDR      0x20000          // 25AA1024(1Mbit)の最大アドレス
#define EEP_NUM           1                // 実装されているEEPROMの数
#define MAX_RECORDS       (8000 * EEP_NUM) // EEPROMに保存できる最大計測ポイント数

//---------------- 配線を変更したら変更する項目 -------------------------------------------
// CTセンサー関連
#define PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION    3  //  Port Digital 3
#define PORT_CH1_RANGE_SELECT_LOW_RESOLUTION     2  //  Port Digital 2
#define PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION   15  //  Port Analog  1
#define PORT_CH2_RANGE_SELECT_LOW_RESOLUTION    14  //  Port Analog  0
#define PORT_ZERO_CROSS 4   // Port Digital  4
#define PORT_CH1        18  // Port Analog   4
#define PORT_CH2        16  // Port Analog   2
#define PORT_REF        17  // Port Analog   3  
#define PORT_BUZZER     19  // Port Analog   5

#ifdef HW_VERSION3
// EEPROM関連
#define EEPROM1_CS     8          // EEPROMのCSピン番号
#define EEPROM2_CS     7          // EEPROMのCSピン番号
#define EEPROM3_CS     6          // EEPROMのCSピン番号
#define EEPROM4_CS     5          // EEPROMのCSピン番号
#endif

//--------------- 方変換の共用体定義 -----------------------------
typedef union {
  unsigned long long_val;
  struct {
    byte LL;  // Lower address
    byte ML;
    byte MH;
    byte HH;  // Higher address
  } byte_val;    
} type_convert_long_byte; 

typedef union {
  unsigned int int_val;
  struct {
    byte L;
    byte H;
  } byte_val;    
} type_convert_int_byte; 

//---------------- 回路定数など -----------------------------------------------------------
#define SELECT          HIGH
#define NOT_SELECT      LOW

#define HIGH_RESOLUTION 0
#define LOW_RESOLUTION  1

#define REF_VOLTS    2.495              // TL431 Typ 2.495V
#define PERIOD  (1000000/FREQ/SAMPLES)  // microsec
const float pi = 3.141592;

int out_volts[SAMPLES];                 // 電流－＞電圧変換回路の出力の計測値
int ac100_sin_volts[SAMPLES];           // AC100V サインカーブ
int ref_volts_10bit;                    // TL431リファレンス電圧のA/D RAW値

//---------------- 変数 -----------------------------------------------------------
boolean zc_circuit_connected;

byte radio_buf_read[32];
byte radio_buf_write[32];
byte serial_buf_read[32];    // シリアル通信受信バッファ
byte serial_buf_index;       // シリアル通信受信バッファインデックス

unsigned long eep_send_addr = 0;   // これから送信すべきアドレス
unsigned long eep_save_addr = 0;   // 次に書き込むべきアドレス

unsigned long watt_ch1;
unsigned long watt_ch2;
unsigned long va_ch1;
unsigned long va_ch2;
unsigned long sampling_counter;

unsigned long sampling_start_at;
unsigned long sampling_end_at;

unsigned long seq_no_save;   // このプログラムを起動してから永久にカウントアップ
unsigned long seq_no_send;   // 送信データのシーケンシャル番号 

bool pc_ready;               // PCとの通信が確立しているときにTrueになる
bool eeprom_full;            // eepromが一杯になったときに計測をとめるためのフラグ
bool wake_up;                // 電源投入後時刻同期が行われていないときtrue
bool ping;                   // trueの時にpingを実行する       

int ch1_range = HIGH_RESOLUTION;   // CH1のレンジ　
int ch2_range = HIGH_RESOLUTION;   // CH2のレンジ　

#ifdef HW_VERSION3
//------------------------------------------------------------------------------------------
// SEQ_No からEEPROMのアドレスに変換する
//------------------------------------------------------------------------------------------
unsigned long seq_to_addr( long seq_no )
{
  return ( seq_no % ((EEP_MAX_ADDR * EEP_NUM) >> 3) ) << 3;
 // return ( seq_no % 0x4000 ) << 3;
}

//------------------------------------------------------------------
// EEPROM disable
//------------------------------------------------------------------
void spi_eeprom_chipselect_HIGH( byte chip_no )
{
  switch( chip_no ) {
    case 1: digitalWrite( EEPROM1_CS, HIGH ); break;
    case 2: digitalWrite( EEPROM2_CS, HIGH ); break;
    case 3: digitalWrite( EEPROM3_CS, HIGH ); break;
    case 4: digitalWrite( EEPROM4_CS, HIGH ); break;
  }
}  
    
//------------------------------------------------------------------
// EEPROM enable
//------------------------------------------------------------------
void spi_eeprom_chipselect_LOW( byte chip_no )
{
  switch( chip_no ) {
    case 1: digitalWrite( EEPROM1_CS, LOW ); break;
    case 2: digitalWrite( EEPROM2_CS, LOW ); break;
    case 3: digitalWrite( EEPROM3_CS, LOW ); break;
    case 4: digitalWrite( EEPROM4_CS, LOW ); break;
  }
}    

//------------------------------------------------------------------------------------------
// EEPROM 25AA1024　に書き込む　
//    実装されている24AA1025の数を判断している
//    複数EEPROMのアドレスを連続アドレスとして扱う
//
// 制約：EEPROMの256バイトページ境界を判断していないのでこれをまたぐ書き込みは不可
// RETURN: 次に書き込むべきアドレス
//------------------------------------------------------------------------------------------
void spi_eeprom_write( unsigned long addr, byte data[], byte len )
{
  byte index;
  type_convert_long_byte eep_addr;
  byte chip_number;
  
  chip_number = (addr / EEP_MAX_ADDR ) + 1;
  spi_eeprom_chipselect_LOW( chip_number );  // eeprom chip select enable 
  SPI.transfer( B00000110 );                 // eeprom enable write operation Instruction
  spi_eeprom_chipselect_HIGH( chip_number ); // eeprom chip select disable     
  
  spi_eeprom_chipselect_LOW( chip_number );  // eeprom chip select enable 
  SPI.transfer( B00000010 );                 // eeprom Write Instruction
  eep_addr.long_val = addr;
  SPI.transfer( eep_addr.byte_val.MH );            
  SPI.transfer( eep_addr.byte_val.ML );
  SPI.transfer( eep_addr.byte_val.LL );
  for( index = 0 ; index < len ; index++ ) {
    SPI.transfer( data[index]);
  }
  spi_eeprom_chipselect_HIGH( chip_number ); // eeprom chip select disable
  delay(6);
}

//------------------------------------------------------------------------------------------
// EEPROM 25AA1024　から読み出す
//    実装されている24AA1025の数を判断している
//    複数EEPROMのアドレスを連続アドレスとして扱う
//
// 制約：EEPROMの256バイトページ境界を判断していないのでこれをまたぐ読み出しは不可
//------------------------------------------------------------------------------------------
void spi_eeprom_read( unsigned long addr, byte data[], byte len )
{
  byte index;
  type_convert_long_byte eep_addr;
  byte chip_number;
  
  chip_number = ( addr / EEP_MAX_ADDR ) + 1;
  
  spi_eeprom_chipselect_LOW( chip_number );  // eeprom chip select enable 
  SPI.transfer( B00000011 );        // eeprom Read Instruction
  eep_addr.long_val = addr;
  SPI.transfer( eep_addr.byte_val.MH );            
  SPI.transfer( eep_addr.byte_val.ML );
  SPI.transfer( eep_addr.byte_val.LL );
  for( index = 0 ; index < len ; index++ ) {
    data[index] = SPI.transfer(0);
  }
  spi_eeprom_chipselect_HIGH( chip_number ); // eeprom chip select disable
}
#endif

//------------------------------------------------------------------------------------------
// ＡＣ１００Ｖサイン波形の生成 
// １／４サイクルでもよいがメモリー不足ではないので１サイクル分生成している
//------------------------------------------------------------------------------------------
void generate_ac100v_sine_curve()
{
  for( int time_index = 0 ; time_index < SAMPLES ; time_index++ ) {
    ac100_sin_volts[time_index] = abs(sin( ( 2 * pi ) / SAMPLES * time_index )) * sqrt(2) * 100; 
  }
}  

//------------------------------------------------------------------------------------------
// １０ビットの計測値を電流値[mA]に換算する
//
// INPUUT:
//   volt     -> 計測対象機器の10ビット計測値
//   ref_volt -> TL431のA/D計測値
//   R_range  -> 計測レンジ
//------------------------------------------------------------------------------------------
float convert_10bit_to_milli_amp( int volt, int ref_volt, int R_range )
{
  return ((float)volt * REF_VOLTS * CT_TURN) / ((float)ref_volt * (float)R_range);
} 

//------------------------------------------------------------------------------------------
// １サイクルぶんをサンプリングする
//
// INPUT:
//   port  -> arduinoアナログポート番号
//   range -> HIGH_RESOLUSION / LOW_RESOLUTION
//------------------------------------------------------------------------------------------
void ad_get( int port, int range )
{
  long time_next_period;
  long period = PERIOD;
 
  if( range == HIGH_RESOLUTION ) {
    if( port == PORT_CH1 ) {
      digitalWrite( PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION, SELECT     );
      digitalWrite( PORT_CH1_RANGE_SELECT_LOW_RESOLUTION,  NOT_SELECT );
    } else {  // PORT_CH2
      digitalWrite( PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION, SELECT     );
      digitalWrite( PORT_CH2_RANGE_SELECT_LOW_RESOLUTION,  NOT_SELECT );
    }
  } else {  // LOW_RESOLUTION
    if( port == PORT_CH1 ) {
      digitalWrite( PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION, NOT_SELECT );
      digitalWrite( PORT_CH1_RANGE_SELECT_LOW_RESOLUTION,  SELECT     );
    } else {  // PORT_CH2
      digitalWrite( PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION, NOT_SELECT );
      digitalWrite( PORT_CH2_RANGE_SELECT_LOW_RESOLUTION,  SELECT     );
    }
  }  
  delay(1); // デジタル回路が動くと電圧が変動したりするので1ms程度待っている。たぶん効果はほとんど無い。気分の問題。

  if( micros() > 4294927296 ) { // for内部でオーバーフローすると大小関係が崩れるのであふれる直前40ms前になったら
    while(micros() < 20000) {}  // オーバーフローするまで待つ
  }
  while(digitalRead(PORT_ZERO_CROSS) == LOW ) {}  // AC100Vの電圧がゼロになるまで待つ
  time_next_period = micros();                    //  プルアップされているのでAC100につながなくてもループを抜ける
  for( int time_index = 0 ; time_index < SAMPLES ; time_index++ ) {
    out_volts[time_index] = analogRead( port );
    time_next_period += period;
    while( micros() < time_next_period ) {
//      digitalWrite( PORT_LED, HIGH );  // LEDが点灯しないならA/D変換速度が間に合っていないので
                                         // サンプル数を小さくすること
    }
  }  
}

//------------------------------------------------------------------------------------------
// リファレンス電圧のＡ／Ｄ変換１０ビット値を得る
// １回Ａ／Ｄ変換すれば済むが、Ａ／Ｄの揺らぎを考慮し１０回取得して平均値を求めている
// INPUT:
//   port  -> arduinoアナログポート番号 = TL431出力
//------------------------------------------------------------------------------------------
float get_ref_volt( int port )
{
  int total = 0;
  
  for( int t = 0 ; t < 10 ; t++ ) {
    delay(1);
    total += analogRead( port );
  }  

  return (float)total / 10;
}

//------------------------------------------------------------------------------------------
// 計測値のオフセット値を求める
// 機器の電流波形は正負均等に触れると見なし、平均値をオフセットとする 値は 大体 512 になる
//------------------------------------------------------------------------------------------
float calc_offset()
{
  long total = 0;

  for(int time_index = 0 ; time_index < SAMPLES ; time_index++ ) {  
    total += (long)out_volts[time_index];
  }

  return (float)total / SAMPLES;
}

//------------------------------------------------------------------------------------------
// １ビットあたりの電流値[mA]を求める
//
// INPUT:
//    range -> HIGH_RESOLUSION / LOW_RESOLUTION
//------------------------------------------------------------------------------------------
float get_amp_div( int port, int range )
{
  if( range == HIGH_RESOLUTION ) {
    if( port = PORT_CH1 ) {
      return (REF_VOLTS * CT_TURN) / ((float)ref_volts_10bit * CH1_R_HIGH_RESOLUTION) * 1000;
    } else { // CH2
      return (REF_VOLTS * CT_TURN) / ((float)ref_volts_10bit * CH2_R_HIGH_RESOLUTION) * 1000;
    }
  } else { // LOW_RESOLUTION
    if( port = PORT_CH1 ) {
      return (REF_VOLTS * CT_TURN) / ((float)ref_volts_10bit * CH1_R_LOW_RESOLUTION) * 1000;
    } else { // CH2
      return (REF_VOLTS * CT_TURN) / ((float)ref_volts_10bit * CH2_R_LOW_RESOLUTION) * 1000;
    }
  }
}

//------------------------------------------------------------------------------------------
// ブザーを０．２秒鳴らす
//------------------------------------------------------------------------------------------  
void beep()
{
  for( int i = 0 ; i < 100 ; i++ ) {
    digitalWrite( PORT_BUZZER, HIGH );
    delay(1);
    digitalWrite( PORT_BUZZER, LOW  );
    delay(1);
  }    
}

//------------------------------------------------------------------------------------------
// ブザーを0.02秒鳴らす
//------------------------------------------------------------------------------------------  
void beep_short()
{
  for( int i = 0 ; i < 10 ; i++ ) {
    digitalWrite( PORT_BUZZER, HIGH );
    delay(1);
    digitalWrite( PORT_BUZZER, LOW  );
    delay(1);
  }    
}

//------------------------------------------------------------------------------------------ 
// 消費電力の計算 ゼロクロス点からサンプリングされた波形であることを前提として電力を計算する
// 返値は精度を確保するため電力の１０倍
//
// RETURN: watt * 10 [W]
//------------------------------------------------------------------------------------------  
unsigned long calc_watt_zc( int port, int range )
{
  float offset;
  float total;
  float temp_volt;
  float amp_div;
  
  offset = calc_offset();

  amp_div = get_amp_div( port, range );
  total = 0.0;
  for( int time_index ; time_index < SAMPLES ; time_index++ ) {  // １サイクルぶん電力を積算
    temp_volt = abs((float)out_volts[time_index] - offset);
    total += (temp_volt * amp_div * (float)ac100_sin_volts[ time_index ] / 100);
  }  

  return (unsigned int)(total / SAMPLES);
}

//------------------------------------------------------------------------------------------
// 電流波形がSIN波になる場合の電流実効値[mA]を求める
// AC100Vを接続していないときはこの計算式を使う。力率１００％なら誤差は無い
// 平均値に波形率1.11を掛けても良いが定義通り 計算式 √avg(Σ計測値^2) で計算している
// 返値は精度を確保するため電力の１０倍
//
// INPUT:
//   out_volts[SAMPLES]   extern
//   port  -> arduino analog port
//   range -> HIGH_RESOLUSION / LOW_RESOLUTION
//
// RETURN: watt * 10 [W]
//------------------------------------------------------------------------------------------
float calc_rms_watt( int port, int range )
{
  float total;
  float temp_volt;
  float amp_div;
  float offset;
  
  offset = calc_offset();
  total = 0.0;
  for( int time_index = 0 ; time_index < SAMPLES ; time_index++ ) {
    temp_volt = abs((float)out_volts[time_index] - offset);
    total += temp_volt * temp_volt;
  }
  amp_div = get_amp_div( port, range );
  return sqrt( total / SAMPLES ) * amp_div * 100 / 100;  
}

//------------------------------------------------------------------------------------------
// 引数と現在値とのミリ秒差を返す　
//   millis()の50日オーバーフローは回避されるが差が3.1日を越えると誤動作する
//------------------------------------------------------------------------------------------
unsigned long millis_diff(unsigned long millis_from)
{
  if( millis_from > 0xF0000000 ) {
    millis_from += 0x10000000;
    return (millis() + 0x10000000) - millis_from;
  } else {
    return millis() - millis_from;
  }
}

#ifdef HW_VERSION3
//------------------------------------------------------------------------------------------
// タイムアウト付き無線受信
//
// timeout: タイムアウト時間[msec]
//------------------------------------------------------------------------------------------
bool radio_recieve( byte buf[], unsigned long timeout )
{
  unsigned long millis_tmp;
  bool millis_overflow = false;
  
 // radio.startListening();
  unsigned long started_waiting_at = millis();
  while(!Mirf.dataReady()){
    if( millis_diff(started_waiting_at) > timeout ) {
      return false;
    }
  }    
  Mirf.getData( buf );
  return true;
}

//------------------------------------------------------------------------------------------
// 無線送信
//------------------------------------------------------------------------------------------
void radio_write( byte buf[] )
{
  Mirf.send( buf );
  
  while( Mirf.isSending() ){}
}
#endif

//------------------------------------------------------------------------------------------
// AC100Vゼロクロス検出回路が接続されているか確認
//------------------------------------------------------------------------------------------
boolean zc_circuit_connect()
{
  unsigned long ms = millis();
    
  while( (millis() - ms) < 200 ) {
    if( digitalRead(PORT_ZERO_CROSS) == LOW ) {
      return true;
    }
  }
  
  beep();
  delay(500);
  beep();
  delay(500);
  
  return false;
}

//------------------------------------------------------------------------------------------
// arduino setup
//------------------------------------------------------------------------------------------  
void setup()
{
  pinMode(PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION, OUTPUT);
  pinMode(PORT_CH1_RANGE_SELECT_LOW_RESOLUTION,  OUTPUT);
  pinMode(PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION, OUTPUT);
  pinMode(PORT_CH2_RANGE_SELECT_LOW_RESOLUTION,  OUTPUT);

  pinMode(PORT_BUZZER, OUTPUT);
  
  digitalWrite( PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION, SELECT     );
  digitalWrite( PORT_CH1_RANGE_SELECT_LOW_RESOLUTION,  NOT_SELECT );
  digitalWrite( PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION, SELECT     );
  digitalWrite( PORT_CH2_RANGE_SELECT_LOW_RESOLUTION,  NOT_SELECT );

#ifdef HW_VERSION3
  pinMode(EEPROM1_CS, OUTPUT);
  pinMode(EEPROM2_CS, OUTPUT);
  pinMode(EEPROM3_CS, OUTPUT);
  pinMode(EEPROM4_CS, OUTPUT);
  digitalWrite( EEPROM1_CS, HIGH );
  digitalWrite( EEPROM2_CS, HIGH );
  digitalWrite( EEPROM3_CS, HIGH );
  digitalWrite( EEPROM4_CS, HIGH );
  SPI.begin();
  // 通信開始
  Mirf.csnPin = 10;
  Mirf.cePin  = 9;
  Mirf.spi = &MirfHardwareSpi;
  Mirf.init();

  Mirf.setRADDR((byte *)"clie1");
  Mirf.setTADDR((byte *)"serv1");
  Mirf.payload = 32;
  Mirf.config();
#else
  lcd.begin(8, 2);
  lcd.println("Watt    ");
  lcd.setCursor(0,1); 
  lcd.println("  Logger");
#endif

#ifdef HW_VERSION1
  zc_circuit_connected = false;
#else
  zc_circuit_connected = zc_circuit_connect();
  generate_ac100v_sine_curve();
#endif

  ref_volts_10bit = get_ref_volt( PORT_REF );

#ifdef HW_VERSION3  
  #ifdef DEBUG  
    Serial.begin(57600);
  #endif
#else
  Serial.begin(57600);
  serial_buf_index = 0;
#endif

  beep();
  delay(1000);
    
  pc_ready     = false;          
  eeprom_full  = false;   
  ping         = false;

  seq_no_save = 0;
  seq_no_send = 0;
  watt_ch1 = 0;
  watt_ch2 = 0;
  va_ch1   = 0;
  va_ch2   = 0;
  sampling_counter = 0;
  
  sampling_start_at = millis();
  sampling_end_at = sampling_start_at + sampling_interval;
}

//------------------------------------------------------------------------------------------
// arduino loop　HardWare Ver3用
//------------------------------------------------------------------------------------------  
void loop_radio()
{
  type_convert_int_byte   val;
  type_convert_long_byte  v;
  unsigned long eep_send_addr_tmp;
  byte eep_buf[8];
  unsigned long eep_addr;
  
  //---------- A/D convert ----------------------------------------
  if( ! eeprom_full ) {
    ad_get( PORT_CH1, ch1_range );
    va_ch1 += calc_rms_watt( PORT_CH1, ch1_range ); 
    if( zc_circuit_connected ) {
      watt_ch1 += calc_watt_zc( PORT_CH1, ch1_range ); 
    }
    ad_get( PORT_CH2, ch2_range );
    va_ch2 += calc_rms_watt( PORT_CH2, ch2_range );
    if( zc_circuit_connected ) {
      watt_ch2 += calc_watt_zc( PORT_CH2, ch2_range );
    }
    sampling_counter++;
  }

  //---------- store EEPROM ---------------------------------------
  if( sampling_end_at < millis() ) {
    if( ! eeprom_full ) {
      val.int_val = (unsigned int)( watt_ch1 / sampling_counter );
      eep_buf[0] = val.byte_val.L;
      eep_buf[1] = val.byte_val.H;
      val.int_val = (unsigned int)( va_ch1 / sampling_counter );
      eep_buf[2] = val.byte_val.L;
      eep_buf[3] = val.byte_val.H;
      val.int_val = (unsigned int)( watt_ch2 / sampling_counter );
      eep_buf[4] = val.byte_val.L;
      eep_buf[5] = val.byte_val.H;
      val.int_val = (unsigned int)( va_ch2 / sampling_counter );
      eep_buf[6] = val.byte_val.L;
      eep_buf[7] = val.byte_val.H;      
      eep_addr = seq_to_addr( seq_no_save ); 
      Serial.print("SAVE ");
      Serial.println(seq_no_save);
      spi_eeprom_write( eep_addr, eep_buf, 8 );
      if( (seq_no_save - seq_no_send ) >= MAX_RECORDS ) {
        eeprom_full = true;
      }
      watt_ch1 = 0;
      watt_ch2 = 0;
      va_ch1   = 0;
      va_ch2   = 0;
      sampling_counter = 0;
    }
    seq_no_save++;
    sampling_start_at = sampling_end_at;
    sampling_end_at = sampling_start_at + sampling_interval;
    ping = true;
  }

  //--------- Ping ------------------------------------
  if( !pc_ready && ping) {
    ping = false;
    radio_buf_write[0] = 'P';
    radio_write( radio_buf_write );
  }

  //---------- PCと接続が確立されているときの処理 --------
  if( pc_ready ) {

    //--------- 電力計測値を送信 -------------------------  
    if( seq_no_send < seq_no_save ) {
      radio_buf_write[0] = 'A';
      v.long_val = seq_no_send;
      radio_buf_write[1] = v.byte_val.LL;
      radio_buf_write[2] = v.byte_val.ML;
      radio_buf_write[3] = v.byte_val.MH;
      radio_buf_write[4] = v.byte_val.HH;

      if( ch1_range == HIGH_RESOLUTION ) {
        radio_buf_write[5] = 'H';
      } else {
        radio_buf_write[5] = 'L';
      }
      if( ch2_range == HIGH_RESOLUTION ) {
        radio_buf_write[6] = 'H';
      } else {
        radio_buf_write[6] = 'L';
      }
      eep_addr = seq_to_addr( seq_no_send ); 
      #ifdef DEBUG
        Serial.print("Send ");
        Serial.println(seq_no_send);
      #endif  
      spi_eeprom_read( eep_addr, &radio_buf_write[7], 8 );
      
      v.long_val = seq_no_save - 1;
      radio_buf_write[15] = v.byte_val.LL;
      radio_buf_write[16] = v.byte_val.ML;
      radio_buf_write[17] = v.byte_val.MH;
      radio_buf_write[18] = v.byte_val.HH;
      
      radio_write( radio_buf_write ); 
      seq_no_send++;
    }  
  } 
  
  //---------- PCからの応答/指令を受信 ---------------------------
  if( radio_recieve( radio_buf_read, 10 ) ) {
    #ifdef DEBUG
      Serial.print( "RECIEVE " );
      Serial.write( radio_buf_read[0] );
    #endif  
    switch( radio_buf_read[0] ) {
      case 'R':
        #ifdef DEBUG
          Serial.println("RANGE CHANGE");
        #endif
        if( radio_buf_read[1] == 'H' ) {
          ch1_range = HIGH_RESOLUTION;
        } else {
          ch1_range = LOW_RESOLUTION;
        }
        if( radio_buf_read[2] == 'H' ) {
          ch2_range = HIGH_RESOLUTION;
        } else {
          ch2_range = LOW_RESOLUTION;
        }
      case 'W':
        pc_ready = false;
        break;
      case 'p':
        switch( radio_buf_read[1] ) {
          case 'R':
            pc_ready = true;
            break;
          case 'N':
            pc_ready = false;
            break;
        }
        break;
      case 'd':
        v.byte_val.LL = radio_buf_read[1];
        v.byte_val.ML = radio_buf_read[2]; 
        v.byte_val.MH = radio_buf_read[3]; 
        v.byte_val.HH = radio_buf_read[4];
        #ifdef DEBUG
          Serial.print("wind request ");
          Serial.println( v.long_val );
        #endif
        unsigned long start_record_no; 
        if( seq_no_save < MAX_RECORDS ) {
          start_record_no = 0;
        } else {
          start_record_no = seq_no_save - MAX_RECORDS;
        }         
        if( start_record_no < v.long_val && v.long_val < seq_no_save ) { 
          seq_no_send = v.long_val;
        }
        break;
      case 'V':
        #ifdef DEBUG
          Serial.print("V command");
        #endif
        radio_buf_write[0] = 'v';
        radio_buf_write[1] = VERSION_NUMBER;
        strncpy( (char *)&radio_buf_write[2], FW_VERSION, 6 );
        val.int_val = (unsigned int)( CH1_R_HIGH_RESOLUTION );
        radio_buf_write[8]  = val.byte_val.L;
        radio_buf_write[9]  = val.byte_val.H;  
        val.int_val = (unsigned int)( CH1_R_LOW_RESOLUTION );
        radio_buf_write[10] = val.byte_val.L;
        radio_buf_write[11] = val.byte_val.H;  
        val.int_val = (unsigned int)( CH2_R_HIGH_RESOLUTION );
        radio_buf_write[12]  = val.byte_val.L;
        radio_buf_write[13]  = val.byte_val.H;  
        val.int_val = (unsigned int)( CH2_R_LOW_RESOLUTION );
        radio_buf_write[14] = val.byte_val.L;
        radio_buf_write[15] = val.byte_val.H;        
        if( zc_circuit_connected ) {
          radio_buf_write[16] = 'Z';
        } else {
          radio_buf_write[16] = '-';
        }
        radio_write( radio_buf_write ); 
        break;
    }
  }
}

//------------------------------------------------------------------------------------------
// arduino loop　HardWare Ver1,2用
//------------------------------------------------------------------------------------------  
void loop_lcd()
{
  byte incomingByte;
  
  //---------- A/D convert ----------------------------------------
  if( ! eeprom_full ) {
    ad_get( PORT_CH1, ch1_range );
    va_ch1 += calc_rms_watt( PORT_CH1, ch1_range ); 
    if( zc_circuit_connected ) {
      watt_ch1 += calc_watt_zc( PORT_CH1, ch1_range ); 
    }
    ad_get( PORT_CH2, ch2_range );
    va_ch2 += calc_rms_watt( PORT_CH2, ch2_range );
    if( zc_circuit_connected ) {
      watt_ch2 += calc_watt_zc( PORT_CH2, ch2_range );
    }
    sampling_counter++;
  }

  //---------- send data ---------------------------------------
  if( sampling_end_at < millis() ) {
    watt_ch1 = watt_ch1 / sampling_counter;
    watt_ch2 = watt_ch2 / sampling_counter;
    va_ch1 = va_ch1 / sampling_counter;
    va_ch2 = va_ch2 / sampling_counter;
    #ifdef HW_VERSION2
   lcd.clear();
    if( zc_circuit_connected ) {  
      lcd.print( "C1 " ); lcd.print( (unsigned int)(watt_ch1 / 10) ); lcd.print( "W" );
      lcd.setCursor(0,1);  
      lcd.print( "C2 " ); lcd.print( (unsigned int)(watt_ch2 / 10) ); lcd.print( "W" );
    } else {
      lcd.print( "C1 " ); lcd.print( (unsigned int)(va_ch1 / 10) ); lcd.print( "W" );
      lcd.setCursor(0,1);  
      lcd.print( "C2 " ); lcd.print( (unsigned int)(va_ch2 / 10) ); lcd.print( "W" );
    } 
    #endif
    #ifdef HW_VERSION1
    lcd.clear();
      lcd.print( "C1 " ); lcd.print( (unsigned int)(va_ch1 / 10) ); lcd.print( "W" );
      lcd.setCursor(0,1);  
      lcd.print( "C2 " ); lcd.print( (unsigned int)(va_ch2 / 10) ); lcd.print( "W" );
    #endif    
    
    if( pc_ready ) {
      Serial.print("A,");
      Serial.print(seq_no_send);
      Serial.print(",");
      if( ch1_range == HIGH_RESOLUTION ) {
        Serial.print("H");
      } else {
        Serial.print("L");
      }
      Serial.print(",");
      if( ch2_range == HIGH_RESOLUTION ) {
        Serial.print("H");
      } else {
        Serial.print("L");
      }
      Serial.print(",");
      Serial.print( (unsigned int)watt_ch1 );
      Serial.print(",");
      Serial.print( (unsigned int)va_ch1   );
      Serial.print(",");
      Serial.print( (unsigned int)watt_ch2 );
      Serial.print(",");
      Serial.print( (unsigned int)va_ch2   );
      Serial.print(",");
      Serial.print(seq_no_send);
      Serial.print("#");
    }
    seq_no_send++;
    watt_ch1 = 0;
    watt_ch2 = 0;
    va_ch1   = 0;
    va_ch2   = 0;
    sampling_counter = 0;
    seq_no_save++;
    sampling_start_at = sampling_end_at;
    sampling_end_at = sampling_start_at + sampling_interval;
  }

  //---------- PCからの応答/指令を受信 ---------------------------
  if( Serial.available() > 0) {
    incomingByte = Serial.read();
    if( incomingByte == '#' ) {
      serial_buf_read[serial_buf_index] = '\0';
      switch(serial_buf_read[0]) {
        case 'S':   // Start
          Serial.print("s#");
          pc_ready = true;
          break;
        case 'W':   // Wait
          Serial.print("w#");
          pc_ready = false;
          break;
        case 'R':
          if( serial_buf_read[1] == 'H' ) {
            ch1_range = HIGH_RESOLUTION;
          } else {
            ch1_range = LOW_RESOLUTION;
          }
          if( serial_buf_read[2] == 'H' ) {
            ch2_range = HIGH_RESOLUTION;
          } else {
            ch2_range = LOW_RESOLUTION;
          }
          break;
        case 'V':
          Serial.print("v,");
          Serial.print( VERSION_NUMBER );        Serial.print(",");
          Serial.print( FW_VERSION );            Serial.print(",");
          Serial.print( CH1_R_HIGH_RESOLUTION ); Serial.print(",");
          Serial.print( CH1_R_LOW_RESOLUTION  ); Serial.print(",");
          Serial.print( CH2_R_HIGH_RESOLUTION ); Serial.print(",");
          Serial.print( CH2_R_LOW_RESOLUTION  ); Serial.print(",");
          if( zc_circuit_connected ) {
            Serial.print( "Z," );
          } else {
            Serial.print( "-," );
          }
          Serial.print( FW_VERSION );            Serial.print("#");
          break;
      }
      serial_buf_index = 0; 
    } else {
      if( incomingByte != 0x0a ) {
        serial_buf_read[serial_buf_index] = incomingByte;
        serial_buf_index++;
      } else {
        serial_buf_index = 0;
      }
    }  
  }
}

//------------------------------------------------------------------------------------------
// arduino loop
//------------------------------------------------------------------------------------------  
void loop()
{
#ifdef HW_VERSION3
  loop_radio();
#else
  loop_lcd();
#endif
}

//------------------------------------------------------------------ E N D ------------------

